"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/model_blender/blender.py
"""

from astropy.io import fits
from stdatamodels import fits_support

from spike.jwstcal.model_blender_rules import make_blender
from spike.jwstcal.model_blender_schemautil import parse_schema
from spike.jwstcal.model_blender_tablebuilder import TableBuilder, table_to_schema


__all__ = ['ModelBlender']


class ModelBlender:
    """
    Class to "blend" metadata from several datamodels.

    Tht output "blended" model will contain:

        - metadata for a combined model
        - a table with metadata of each datamodel

    Input models can be added to the blender using `ModelBlender.accumulate`
    and the output/combined model updated using `ModelBlender.finalize_model`.

    All input/accumulated models must be of the same type.

    >>> blender = ModelBlender()
    >>> blender.accumulate(input_model_a)  # doctest: +SKIP
    >>> blender.accumulate(input_model_b)  # doctest: +SKIP
    >>> blender.finalize_model(combined_model)  # doctest: +SKIP

    """
    def __init__(self, blend_ignore_attrs=None):
        """
        Create a new `ModelBlender`.

        Parameters
        ----------
        blend_ignore_attrs: list or None
            A list of metadata attributes to ignore during blending.
            These attributes will not be set on the output/combined.
            These attributes must be strings containing the dotted
            path of each attribute (for example "meta.filename").
            (Note that "meta.wcs" will always be ignored).
        """
        self._model_type = None
        self._first_header_meta = None
        self._blenders = None
        self._table_builder = None
        self._blend_ignore_attrs = ['meta.wcs']
        if blend_ignore_attrs is not None:
            self._blend_ignore_attrs.extend(blend_ignore_attrs)

    def accumulate(self, model):
        """
        Blend metadata for model.

        Process model adding its metadata to the blended
        metadata and the metadata table.

        Parameters
        ----------
        model: `jwst.datamodels.JwstDataModel`
            The datamodel to blend.
        """
        if self._first_header_meta is None:
            self._model_type = type(model)
            # search the schema for other metadata to "blend" and to add to the table
            attr_to_columns, attr_to_blend_rules, schema_ignores  = parse_schema(model.schema)

            # update ignores list for items in schema that can't be blended
            self._blend_ignore_attrs.extend(schema_ignores)

            # capture the entire contents of the first model metadata
            self._first_header_meta = {}
            for attr, v in model.to_flat_dict(include_arrays=False).items():
                if not attr.startswith('meta'):
                    continue
                if any((attr.startswith(i) for i in self._blend_ignore_attrs)):
                    continue
                self._first_header_meta[attr] = v

            # make "blenders" for the metadata with special rules
            self._blenders = {}
            for attr, rule in attr_to_blend_rules.items():
                if rule == 'first':
                    continue
                if any((attr.startswith(i) for i in self._blend_ignore_attrs)):
                    continue
                self._blenders[attr] = make_blender(rule)

            # make a table builder using the mapping from the schema
            self._table_builder = TableBuilder(attr_to_columns)
        else:
            if type(model) != self._model_type:  # noqa: E721
                raise ValueError(
                    f"model of type {type(model)} does not match previous type({self._model_type}). "
                    "ModelBlender only supports blending models of the same model type."
                )

        # convert the model to a flat header
        header = model.to_flat_dict(include_arrays=False)

        # add the header to the table
        self._table_builder.header_to_row(header)

        # and perform any special blending
        for attr, blender in self._blenders.items():
            if attr in header:
                blender.accumulate(header[attr])

    def _finalize_metadata(self):
        # start with the entire contents of the first model
        meta = self._first_header_meta.copy()
        for attr, blender in self._blenders.items():
            meta[attr] = blender.finalize()
        return meta

    def _finalize_table(self):
        return self._table_builder.build_table()

    def finalize_model(self, model):
        """
        Update model with the blend results.

        Add blended metadata and the accumulated metadata table to
        the provided datamodel. The update process involves:

            - setting the model metadata to the blended metadata values
            - adding an "hdrtab" attribute (containing the metadata table)
            - updating the model schema to save "hdrtab"

        The provided model will be updated in-place.

        Parameters
        ----------
        model: `jwst.datamodels.JwstDataModel`
            A datamodel that will have its metadata set
            to the blended metadata and have the metadata
            table assigned to the "hdrtab" attribute.
        """
        # update metadata of the output model based on the results
        # of the "blenders"
        for attr, val in self._finalize_metadata().items():
            try:
                model[attr] = val
            except KeyError:
                # Ignore keys that are in the asdf tree but not in the schema
                pass

        # patch the table into the output model and the schema
        table = self._finalize_table()
        schema = table_to_schema(table)
        model.add_schema_entry('hdrtab', schema)

        # because astropy will silently mangle boolean columns on write
        # we roundtrip the data through a BinTableHDU here to allow
        # stdatamodels to correct the data in a way that won't result in mangling
        model.hdrtab = fits_support.from_fits_hdu(fits.BinTableHDU.from_columns(table), schema)


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/entry_points.py
"""

import warnings
from collections import namedtuple

import importlib_metadata

STEPS_GROUP = "stpipe.steps"


StepInfo = namedtuple(
    "StepInfo",
    ["class_name", "class_alias", "is_pipeline", "package_name", "package_version"],
)


def get_steps():
    """
    Get the list of steps registered with stpipe's entry point group.  Each entry
    point is expected to return a list of tuples, where the first tuple element
    is a fully-qualified Step subclass name, the second element is an optional
    class alias, and the third is a bool indicating whether the class is to be
    listed as a pipeline in the CLI output.

    Returns
    -------
    list of StepInfo
    """
    steps = []

    for entry_point in importlib_metadata.entry_points(group=STEPS_GROUP):
        package_name = entry_point.dist.name
        package_version = entry_point.dist.version
        package_steps = []

        try:
            elements = entry_point.load()()
            package_steps = [
                StepInfo(*element, package_name, package_version)
                for element in elements
            ]

        except Exception as e:
            warnings.warn(
                f"{STEPS_GROUP} plugin from package {package_name}=={package_version} "
                "failed to load:\n\n"
                f"{e.__class__.__name__}: {e}",
                stacklevel=2,
            )

        steps.extend(package_steps)

    return steps


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/stpipe/utilities.py
"""

import inspect
import logging
from importlib import import_module
from pkgutil import walk_packages

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Step classes that are not user-api steps
NON_STEPS = [
    "EngDBLogStep",
    "FunctionWrapper",
    "RomancalPipeline",
    "RomanPipeline",
    "ExposurePipeline",
    "MosaicPipeline",
    "RomanStep",
    "Pipeline",
    "Step",
    "SystemCall",
]


def all_steps():
    """List all classes subclassed from Step

    Returns
    -------
    steps : dict
        Key is the classname, value is the class
    """
    from spike.romancal.stpipe_core import RomanStep as Step

    romancal = import_module("romancal")

    steps = {}
    for module in load_sub_modules(romancal):
        more_steps = {
            klass_name: klass
            for klass_name, klass in inspect.getmembers(
                module, lambda o: inspect.isclass(o) and issubclass(o, Step)
            )
            if klass_name not in NON_STEPS
        }
        steps.update(more_steps)

    return steps


def load_sub_modules(module):
    """
    Recursively loads all submodules of a module (this is not a local import).

    Parameters
    ----------
    module : module
        A python module to walk, load

    Returns
    -------
    generator
        A generator of all submodules of module recursively until no more sub modules are found
    """

    for package_info in walk_packages(module.__path__):
        if package_info.module_finder.path.startswith(module.__path__[0]):
            package = import_module(f"{module.__name__}.{package_info.name}")

            if package_info.ispkg:
                yield from load_sub_modules(package)

            yield package


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/registry.py
"""

import importlib.util
from inspect import (
    getmembers,
    isclass,
    isfunction,
    ismethod,
    ismodule
)
import logging
from os.path import (
    basename,
    expanduser,
    expandvars,
)
from enum import EnumMeta

from spike.jwstcal import libpath
from spike.jwstcal.associations_exceptions import (
    AssociationError,
    AssociationNotValidError
)
from spike.jwstcal.associations_lib_callback_registry import CallbackRegistry

__all__ = [
    'AssociationRegistry',
    'RegistryMarker'
]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Library files
_ASN_RULE = 'association_rules.py'


class AssociationRegistry(dict):
    """The available associations

    Parameters
    ----------
    definition_files : [str,]
        The files to find the association definitions in.

    include_default : bool
        True to include the default definitions.

    global_constraints : Constraint
        Constraints to be added to each rule.

    name : str
        An identifying string, used to prefix rule names.

    include_bases : bool
        If True, include base classes not considered
        rules.

    Notes
    -----
    The general workflow is as follows:

        * Create the registry
            >>> from jwst.associations.registry import AssociationRegistry
            >>> registry = AssociationRegistry()

        * Create associations from an item
            >>> associations, reprocess = registry.match(item) # doctest: +SKIP

        * Finalize the associations
            >>> final_asns = registry.callback.reduce('finalize', associations) # doctest: +SKIP

    In practice, this is one step in a larger loop over all items to
    be associated. This does not account for adding items to already
    existing associations. See :py:func:`~jwst.associations.generate` for more information.
    """

    def __init__(self,
                 definition_files=None,
                 include_default=True,
                 global_constraints=None,
                 name=None,
                 include_bases=False):
        super().__init__()

        # Generate a UUID for this instance. Used to modify rule
        # names.
        self.name = name

        # Callback registry
        self.callback = CallbackRegistry()

        # Precache the set of rules
        self._rule_set = set()

        if definition_files is None:
            definition_files = []
        if include_default:
            definition_files.insert(0, libpath(_ASN_RULE))
        if len(definition_files) <= 0:
            raise AssociationError('No rule definition files specified.')

        self.schemas = []
        self.Utility = type('Utility', (object,), {})
        for fname in definition_files:
            module = import_from_file(fname)
            self.populate(
                module,
                global_constraints=global_constraints,
                include_bases=include_bases
            )

    @property
    def rule_set(self):
        """Rules within the Registry"""
        return self._rule_set

    def match(self, item, version_id=None, allow=None, ignore=None):
        """See if item belongs to any of the associations defined.

        Parameters
        ----------
        item : dict
            An item, like from a Pool, to find associations for.

        version_id : str
            If specified, a string appended to association names.
            If None, nothing is used.

        allow : [type(Association), ...]
            List of rules to allow to be matched. If None, all
            available rules will be used.

        ignore : list
            A list of associations to ignore when looking for a match.
            Intended to ensure that already created associations
            are not re-created.

        Returns
        -------
        (associations, reprocess_list): 2-tuple
            associations : [association,...]
                List of associations item belongs to. Empty if none match
            reprocess_list : [AssociationReprocess, ...]
                List of reprocess events.
        """
        if allow is None:
            allow = self.rule_set
        if ignore is None:
            ignore = []
        associations = []
        process_list = []
        for name, rule in self.items():
            if rule not in ignore and rule in allow:
                asn, reprocess = rule.create(item, version_id)
                process_list.extend(reprocess)
                if asn is not None:
                    associations.append(asn)
        return associations, process_list

    def validate(self, association):
        """Validate a given association

        Parameters
        ----------
        association : association-like
            The data to validate

        Returns
        -------
        rules : list
            List of rules that validated

        Raises
        ------
        AssociationNotValidError
            Association did not validate
        """

        # Change rule validation from an exception
        # to a boolean
        def is_valid(rule, association):
            try:
                rule.validate(association)
            except AssociationNotValidError:
                return False
            else:
                return True

        results = [
            rule
            for rule_name, rule in self.items()
            if is_valid(rule, association)
        ]

        if len(results) == 0:
            raise AssociationNotValidError(
                'Structure did not validate: "{}"'.format(association)
            )
        return results

    def load(
            self,
            serialized,
            format=None,
            validate=True,
            first=True,
            **kwargs
    ):
        """Load a previously serialized association

        Parameters
        ----------
        serialized : object
            The serialized form of the association.

        format : str or None
            The format to force. If None, try all available.

        validate : bool
            Validate against the class' defined schema, if any.

        first : bool
            A serialization potentially matches many rules.
            Only return the first succesful load.

        kwargs : dict
            Other arguments to pass to the `load` method

        Returns
        -------
        The Association object, or the list of association objects.

        Raises
        ------
        AssociationError
            Cannot create or validate the association.
        """
        results = []
        for rule_name, rule in self.items():
            try:
                results.append(
                    rule.load(
                        serialized,
                        format=format,
                        validate=validate,
                        **kwargs
                    )
                )
            except (AssociationError, AttributeError) as err:
                lasterr = err
                continue
            if first:
                break
        if len(results) == 0:
            raise lasterr
        if first:
            return results[0]
        else:
            return results

    def populate(self,
                 module,
                 global_constraints=None,
                 include_bases=None
                 ):
        """Parse out all rules and callbacks in a module and add them to the registry

        Parameters
        ----------
        module : module
            The module, and all submodules, to be parsed.
        """
        for name, obj in get_marked(module, include_bases=include_bases):

            # Add rules.
            if include_bases or obj._asnreg_role == 'rule':
                try:
                    self.add_rule(name, obj, global_constraints=global_constraints)
                except TypeError:
                    logger.debug(f'Could not add object {obj} as a rule due to TypeError')
                continue

            # Add callbacks
            if obj._asnreg_role == 'callback':
                for event in obj._asnreg_events:
                    self.callback.add(event, obj)
                    continue

            # Add schema
            if obj._asnreg_role == 'schema':
                self.schemas.append(obj._asnreg_schema)
                continue

            # Add utility classes
            if obj._asnreg_role == 'utility':
                self.Utility = type(
                    'Utility',
                    (obj, self.Utility),
                    {}
                )

    def add_rule(self, name, obj, global_constraints=None):
        """Add object as rule to registry

        Parameters
        ----------
        name : str
            Name of the object

        obj : object
            The object to be considered a rule

        global_constraints : dict
            The global constraints to attach to the rule.
        """
        try:
            rule_name = '_'.join([self.name, name])
        except TypeError:
            rule_name = name
        if not valid_class(obj):
            raise TypeError(f'Object cannot be used as rule: {obj}')
        rule = type(rule_name, (obj,), {})
        rule.GLOBAL_CONSTRAINT = global_constraints
        rule.registry = self
        self.__setitem__(rule_name, rule)
        self._rule_set.add(rule)


class RegistryMarker:
    """Mark rules, callbacks, and modules for inclusion into a registry"""

    class Schema:
        def __init__(self, obj):
            self._asnreg_role = 'schema'
            self._asnreg_schema = obj
            RegistryMarker.mark(self)

        @property
        def schema(self):
            return self._asnreg_schema

    @staticmethod
    def mark(obj):
        """Mark that an object should be part of the registry

        Parameters
        ----------
        obj : object
            The object to mark

        Returns
        -------
        obj
            Object that has been marked. Returned to enable
            use as a decorator.

        Notes
        -----
        The following attributes are added to the object:

        - _asnreg_mark : True
              Attribute added to object and is set to True

        - _asnreg_role : str or None
              If not already assigned, the role is left
              unspecified using None.

        """
        obj._asnreg_marked = True
        obj._asnreg_role = getattr(obj, '_asnreg_role', None)
        return obj

    @staticmethod
    def rule(obj):
        """Mark object as rule

        Parameters
        ----------
        obj : object
            The object that should be treated as a rule

        Returns
        -------
        obj : object
            Return object to enable use as a decorator.

        Notes
        -----
        The following attributes are added to the object:

        - _asnreg_role : 'rule'
              Attributed added to object and set to `rule`

        - _asnreg_mark : True
              Attributed added to object and set to True
        """
        obj._asnreg_role = 'rule'
        RegistryMarker.mark(obj)
        return obj

    @staticmethod
    def callback(event):
        """Mark object as a callback for an event

        Parameters
        ----------
        event : str
            Event this is a callback for.

        obj : func
            Function, or any callable, to be called
            when the corresponding event is triggered.

        Returns
        -------
        func
            Function to use as a decorator for the object to be marked.

        Notes
        -----
        The following attributes are added to the object:

        - _asnreg_role : 'callback'
              The role the object as been assigned.
        - _asnreg_events : [event[, ...]]
              The events this callable object is a callback for.
        - _asnreg_mark : True
              Indicated that the object has been marked.
        """
        def decorator(func):
            try:
                events = func._asnreg_events
            except AttributeError:
                events = list()
            events.append(event)
            RegistryMarker.mark(func)
            func._asnreg_role = 'callback'
            func._asnreg_events = events
            return func
        return decorator

    @staticmethod
    def schema(filename):
        """Mark a file as a schema source"""
        schema = RegistryMarker.Schema(filename)
        return schema

    @staticmethod
    def utility(class_obj):
        """Mark the class as a Utility class"""
        class_obj._asnreg_role = 'utility'
        RegistryMarker.mark(class_obj)
        return class_obj

    @staticmethod
    def is_marked(obj):
        """Has an objected been marked?"""
        return hasattr(obj, '_asnreg_marked')


# Utilities
def import_from_file(filename):
    """Import a file as a module

    Parameters
    ---------
    filename : str
        The file to import

    Returns
    -------
    module : python module
        The imported module
    """
    path = expandvars(expanduser(filename))
    module_name = basename(path).split('.')[0]
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def get_marked(module, predicate=None, include_bases=False):
    """Recursively get all executable objects

    Parameters
    ----------
    module : python module
        The module to examine

    predicate : bool func(object)
        Determinant of what gets returned.
        If None, all object types are examined

    include_bases : bool
        If True, include base classes not considered
        rules.

    Returns
    -------
    class object : generator
        A generator that will yield all class members in the module.
    """
    def is_method(obj):
        return isfunction(obj) or ismethod(obj)

    for name, obj in getmembers(module, predicate):
        if isclass(obj):
            for sub_name, sub_obj in get_marked(obj, predicate=is_method):
                yield sub_name, sub_obj
            if RegistryMarker.is_marked(obj) or include_bases:
                yield name, obj
        elif RegistryMarker.is_marked(obj):
            if ismodule(obj):
                for sub_name, sub_obj in get_marked(
                        obj, predicate=predicate, include_bases=include_bases
                ):
                    yield sub_name, sub_obj
            else:
                yield name, obj


def valid_class(obj):
    """Verify if a given object could be used as a rule class

    Parameters
    ------------
    obj : obj
        Object to check

    Returns
    --------
    is_valid : bool
        True if the object could be considered a rule class
    """
    is_valid = type(obj) is not EnumMeta and isclass(obj)
    return is_valid


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/pipeline.py
"""

"""
Pipeline
"""

from collections.abc import Sequence
from os.path import dirname, join
from typing import ClassVar

from astropy.extern.configobj.configobj import ConfigObj, Section

from spike.stpipe import config_parser, crds_client, log
from spike.stpipe.library import AbstractModelLibrary
from spike.stpipe.step import Step, get_disable_crds_steppars
from spike.stpipe.utilities import _not_set

# For classmethods, the logger to use is the
# delegator, since the pipeline has not yet been instantiated.
logger = log.delegator.log


class Pipeline(Step):
    """
    A Pipeline is a way of combining a number of steps together.
    """

    # Configuration
    spec = """
    """
    # A set of steps used in the Pipeline.  Should be overridden by
    # the subclass.
    step_defs: ClassVar = {}

    def __init__(self, *args, **kwargs):
        """
        See `~stpipe.step.Step` for the parameters.
        """
        Step.__init__(self, *args, **kwargs)

        # Configure all of the steps
        for key, val in self.step_defs.items():
            cfg = self.steps.get(key)
            if cfg is not None:
                new_step = val.from_config_section(
                    cfg,
                    parent=self,
                    name=key,
                    config_file=self.config_file,
                )
            else:
                new_step = val(
                    key,
                    parent=self,
                    config_file=self.config_file,
                    **kwargs.get(key, {}),
                )

            setattr(self, key, new_step)

    @property
    def reference_file_types(self):
        """Collect the list of all reftypes for child Steps that are not skipped.
        Overridden reftypes are included but handled normally later by the
        Pipeline version of the get_ref_override() method defined below.
        """
        return [
            reftype
            for step in self._unskipped_steps
            for reftype in step.reference_file_types
        ]

    @property
    def _unskipped_steps(self):
        """Return a list of the unskipped Step objects launched by `self`.

        Steps are also excluded if `Step.prefetch_references` is False.
        """
        return [
            getattr(self, name)
            for name in self.step_defs
            if (
                not getattr(self, name).skip and getattr(self, name).prefetch_references
            )
        ]

    def get_ref_override(self, reference_file_type):
        """Return any override for ``reference_file_type`` for any of the steps in
        Pipeline ``self``.  OVERRIDES Step.

        Returns
        -------
        override_filepath or None.

        """
        for step in self._unskipped_steps:
            override = step.get_ref_override(reference_file_type)
            if override is not None:
                return override
        return None

    @classmethod
    def merge_config(cls, config, config_file):
        steps = config.get("steps", {})

        # Configure all of the steps
        for key in cls.step_defs:
            cfg = steps.get(key)
            if cfg is not None:
                # If a config_file is specified, load those values and
                # then override them with our values.
                if cfg.get("config_file"):
                    cfg2 = config_parser.load_config_file(
                        join(dirname(config_file or ""), cfg.get("config_file"))
                    )
                    del cfg["config_file"]
                    config_parser.merge_config(cfg2, cfg)
                    steps[key] = cfg2
        return config

    @classmethod
    def load_spec_file(cls, preserve_comments=_not_set):
        spec = config_parser.get_merged_spec_file(
            cls, preserve_comments=preserve_comments
        )

        spec["steps"] = Section(spec, spec.depth + 1, spec.main, name="steps")
        steps = spec["steps"]
        for key, val in cls.step_defs.items():
            if not issubclass(val, Step):
                raise TypeError(f"Entry {key!r} in step_defs is not a Step subclass")
            stepspec = val.load_spec_file(preserve_comments=preserve_comments)
            steps[key] = Section(steps, steps.depth + 1, steps.main, name=key)

            config_parser.merge_config(steps[key], stepspec)

            # Also add a key that can be used to specify an external
            # config_file
            step = spec["steps"][key]
            step["config_file"] = "string(default=None)"
            step["name"] = "string(default='')"
            step["class"] = "string(default='')"

        return spec

    @classmethod
    def get_config_from_reference(cls, dataset, disable=None, crds_observatory=None):
        """Retrieve step parameters from reference database

        Parameters
        ----------
        cls : `stpipe.step.Step`
            Either a class or instance of a class derived
            from `Step`.

        dataset : `stpipe.datamodel.AbstractDataModel`
            A model of the input file.  Metadata on this input file will
            be used by the CRDS "bestref" algorithm to obtain a reference
            file.

        disable: bool or None
            Do not retrieve parameters from CRDS. If None, check global settings.

        crds_observatory : str
            Observatory name ('jwst' or 'roman').

        Returns
        -------
        step_parameters : configobj
            The parameters as retrieved from CRDS. If there is an issue, log as such
            and return an empty config obj.
        """
        reftype = cls.get_config_reftype()
        refcfg = ConfigObj()
        refcfg["steps"] = Section(refcfg, refcfg.depth + 1, refcfg.main, name="steps")

        # Check if retrieval should be attempted.
        if disable is None:
            disable = get_disable_crds_steppars()
        if disable:
            logger.debug(
                "%s: CRDS parameter reference retrieval disabled.", reftype.upper()
            )
            return refcfg

        logger.debug("Retrieving all substep parameters from CRDS")
        #
        # Iterate over the steps in the pipeline
        with cls._datamodels_open(dataset, asn_n_members=1) as model:
            if isinstance(model, Sequence):
                crds_parameters = model._models[0].get_crds_parameters()
                crds_observatory = model.crds_observatory
            else:
                crds_parameters = model.get_crds_parameters()
                crds_observatory = model.crds_observatory

        for cal_step in cls.step_defs.keys():
            cal_step_class = cls.step_defs[cal_step]
            refcfg["steps"][cal_step] = cal_step_class.get_config_from_reference(
                crds_parameters,
                crds_observatory=crds_observatory,
            )
        #
        # Now merge any config parameters from the step cfg file
        logger.debug("Retrieving pipeline %s parameters from CRDS", reftype.upper())
        try:
            ref_file = crds_client.get_reference_file(
                crds_parameters,
                reftype,
                crds_observatory,
            )
        except (AttributeError, crds_client.CrdsError):
            logger.debug("%s: No parameters found", reftype.upper())
        else:
            if ref_file != "N/A":
                logger.info("%s parameters found: %s", reftype.upper(), ref_file)
                refcfg = cls.merge_pipeline_config(refcfg, ref_file)
            else:
                logger.debug("No %s reference files found.", reftype.upper())

        return refcfg

    @classmethod
    def merge_pipeline_config(cls, refcfg, ref_file):
        """
        Merge the config parameters from a pipeline config reference file into the
        config obtained from each step

        Parameters
        ----------
        cls : jwst.stpipe.pipeline.Pipeline class
            The pipeline class

        refcfg : ConfigObj object
            The ConfigObj created from crds cfg files from each of the steps
            in the pipeline

        ref_file : string
            The name of the pipeline crds step config file

        Returns
        -------
        ConfigObj of the merged parameters, with those from the pipeline cfg having
        precedence over those from the individual steps
        """

        pipeline_cfg = config_parser.load_config_file(ref_file)
        config_parser.merge_config(refcfg, pipeline_cfg)
        return refcfg

    def _precache_references(self, input_file):
        """
        Precache all of the expected reference files before the Step's
        process method is called.

        Handles opening `input_file` as a model if it is a filename.

        input_file:  filename, model container, or model

        Returns
        -------
        None
        """
        try:
            with self.open_model(
                input_file, asn_n_members=1, asn_exptypes=["science"]
            ) as model:
                self._precache_references_opened(model)
        except (ValueError, TypeError, OSError):
            self.log.info("First argument %s does not appear to be a model", input_file)

    def _precache_references_opened(self, model_or_container):
        """Pre-fetches references for `model_or_container`.

        Handles recursive pre-fetches for any models inside a container,
        or just a single model.

        Assumes model_or_container is an open model or container object,
        not a filename.

        No garbage collection.
        """
        if isinstance(model_or_container, Sequence):
            # recurse on each contained model
            for contained_model in model_or_container:
                self._precache_references_opened(contained_model)
        elif isinstance(model_or_container, AbstractModelLibrary):
            with model_or_container:
                for i, model in enumerate(model_or_container):
                    self._precache_references_impl(model)
                    model_or_container.shelve(model, i, modify=False)
        else:
            # precache a single model object
            self._precache_references_impl(model_or_container)

    def _precache_references_impl(self, model):
        """Given open data `model`,  determine and cache reference files for
        any reference types which are not overridden on the command line.

        Verify that all CRDS and overridden reference files are readable.

        Parameters
        ----------
        model :  `DataModel`
            Only a `DataModel` instance is allowed.
            Cannot be a filename, Sequence, etc.
        """
        ovr_refs = {
            reftype: self.get_ref_override(reftype)
            for reftype in self.reference_file_types
            if self.get_ref_override(reftype) is not None
        }

        fetch_types = sorted(set(self.reference_file_types) - set(ovr_refs.keys()))

        self.log.info(
            "Prefetching reference files for dataset: %r reftypes = %r",
            model.meta.filename,
            fetch_types,
        )
        crds_refs = crds_client.get_multiple_reference_paths(
            model.get_crds_parameters(), fetch_types, model.crds_observatory
        )

        ref_path_map = dict(list(crds_refs.items()) + list(ovr_refs.items()))

        for reftype, refpath in sorted(ref_path_map.items()):
            how = "Override" if reftype in ovr_refs else "Prefetch"
            self.log.info(
                "%s for %s reference file is '%s'.", how, reftype.upper(), refpath
            )
            crds_client.check_reference_open(refpath)

    def get_pars(self, full_spec=True):
        """Retrieve the configuration parameters of a pipeline

        Parameters are retrieved for the pipeline and all of its
        component steps.

        Parameters
        ----------
        full_spec : bool
            Return all parameters, including parent-specified parameters.
            If `False`, return only parameters specific to the pipeline
            and steps.

        Returns
        -------
        pars : dict
            Keys are the parameters and values are the values.
        """
        pars = super().get_pars(full_spec=full_spec)
        pars["steps"] = {}
        for step_name in self.step_defs.keys():
            pars["steps"][step_name] = getattr(self, step_name).get_pars(
                full_spec=full_spec
            )
        return pars


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/lib/catalog_utils.py
"""

import re
from os.path import split, splitext, join, abspath, expanduser
from collections import namedtuple


def replace_suffix_ext(filename, old_suffix_list, new_suffix,
                       output_ext='ecsv', output_dir=None):
    """
    Replace the suffix and extension of a filename.

    If the last suffix in the input filename is in the
    ``old_suffix_list`` then it is replaced by the ``new_suffix``.
    Otherwise, the ``new_suffix`` is appended to the input filename.

    Parameters
    ----------
    filename : str
        The filename to modify.

    old_suffix_list : list of str
        The list of filename suffixes that will be replaced.

    new_suffix : str
        The new filename suffix.

    output_ext : str, optional
        The extension of the output filename.  The default is 'ecsv'.

    output_dir : str or `None`, optional
        The output directory name.  If `None` then the current directory
        will be used.

    Returns
    -------
    result : str
        The modified filename.

    Examples
    --------
    >>> from jwst.lib.catalog_utils import replace_suffix_ext
    >>> replace_suffix_ext('jw12345_nrca_i2d.fits', ['i2d'], 'cat')
        'jw12345_nrca_cat.ecsv'

    >>> replace_suffix_ext('jw12345_nrca_cal.fits', ['i2d'], 'cat')
        'jw12345_nrca_cal_cat.ecsv'

    >>> replace_suffix_ext('my_custom_file.fits', ['i2d'], 'cat')
        'my_custom_file_cat.ecsv'

    >>> old_suffixes = ['calints', 'crfints']
    >>> replace_suffix_ext('jw12345_nrca_calints.fits', old_suffixes, 'phot')
        'jw12345_nrca_phot.ecsv'
    >>> replace_suffix_ext('jw12345_nrca_crfints.fits', old_suffixes, 'phot')
        'jw12345_nrca_phot.ecsv'

    >>> replace_suffix_ext('jw12345_nrca_i2d.fits', ['i2d'], 'cat',
    ...                    output_dir='/jwst/my_catalogs')
        '/jwst/my_catalogs/jw12345_nrca_cat.ecsv'
    """

    path, filename = split(filename)
    name, ext = splitext(filename)
    remove_suffix = '^(.+?)(_(' + '|'.join(old_suffix_list) + '))?$'
    match = re.match(remove_suffix, name)
    name = match.group(1)

    output_path = '{0}_{1}.{2}'.format(name, new_suffix, output_ext)
    if output_dir is not None:
        output_path = abspath(expanduser(join(output_dir, output_path)))

    return output_path


class SkyObject(namedtuple('SkyObject', ("label",
                                         "xcentroid",
                                         "ycentroid",
                                         "sky_centroid",
                                         "isophotal_abmag",
                                         "isophotal_abmag_err",
                                         "sky_bbox_ll",
                                         "sky_bbox_lr",
                                         "sky_bbox_ul",
                                         "sky_bbox_ur",
                                         "is_extended",
                                         ), rename=False)):

    """
    Sky Object container for WFSS catalog information.

    This is a convenience object for storing the catalog information
    as a named tuple. The object has explicit fields to guard for changing
    column locations in the catalog file that's read. Callers should
    validate for the minimum fields they require. This is currently populated
    for the minimum information needed by the WFSS modes in nircam and niriss.

    Parameters
    ----------
    label : int
        source identified
    xcentroid : float
        x center of object in pixels
    ycentroid : float
        y center of object in pixels
    sky_centroid: `~astropy.coordinates.SkyCoord`
        ra and dec of the center of the object
    isophotal_abmag : float
        AB Magnitude of object
    isophotal_abmag_err : float
        Error on the AB magnitude
    sky_bbox_ll : `~astropy.coordinates.SkyCoord`
        Lower left corner of the minimum bounding box
    sky_bbox_lr : `~astropy.coordinates.SkyCoord`
        Lower right corder of the minimum bounding box
    sky_bbox_ul : `~astropy.coordinates.SkyCoord`
        Upper left corner of the minimum bounding box
    sky_bbox_ur : `~astropy.coordinates.SkyCoord`
        Upper right corner of the minimum bounding box
    is_extended : bool
        Flag indicating if the object is extended
    """

    __slots__ = ()  # prevent instance dictionary creation for lower mem

    def __new__(cls, label=None,
                xcentroid=None,
                ycentroid=None,
                sky_centroid=None,
                isophotal_abmag=None,
                isophotal_abmag_err=None,
                sky_bbox_ll=None,
                sky_bbox_lr=None,
                sky_bbox_ul=None,
                sky_bbox_ur=None,
                is_extended=None,):

        return super(SkyObject, cls).__new__(cls,
                                             label=label,
                                             xcentroid=xcentroid,
                                             ycentroid=ycentroid,
                                             sky_centroid=sky_centroid,
                                             isophotal_abmag=isophotal_abmag,
                                             isophotal_abmag_err=isophotal_abmag_err,
                                             sky_bbox_ll=sky_bbox_ll,
                                             sky_bbox_lr=sky_bbox_lr,
                                             sky_bbox_ul=sky_bbox_ul,
                                             sky_bbox_ur=sky_bbox_ur,
                                             is_extended=is_extended
                                             )

    def __str__(self):
        """Return a pretty print for the object information."""
        return ("label: {0}\n"
                "xcentroid: {1}\n"
                "ycentroid: {2}\n"
                "sky_centroid: {3}\n"
                "isophotal_abmag: {4}\n"
                "isophotal_abmag_err: {5}\n"
                "sky_bbox_ll: {6}\n"
                "sky_bbox_lr: {7}\n"
                "sky_bbox_ul: {8}\n"
                "sky_bbox_ur: {9}\n"
                "is_extended: {10}"
                .format(self.label,
                        self.xcentroid,
                        self.ycentroid,
                        str(self.sky_centroid),
                        self.isophotal_abmag,
                        self.isophotal_abmag_err,
                        str(self.sky_bbox_ll),
                        str(self.sky_bbox_lr),
                        str(self.sky_bbox_ul),
                        str(self.sky_bbox_ur),
                        str(self.is_extended)
                        )
                )